/*
	AssimpRead.cpp
	It reads mesh files by Asssimp and builds SIrrMeshBuild structure.

    MUTATE Systems 2018 (http://mutate.jp)
*/

#include <iostream>
#include <vector>
#include "AssimpRead.h"
#include "IrrMeshBuild.h"

#include <assimp/scene.h>          // Output data structure
#include <assimp/postprocess.h>    // Post processing flags
#include <assimp/Importer.hpp>

#pragma comment(lib, "assimp-vc110-mt.lib")

using irr::core::vector2df;
using irr::core::vector3df;
using irr::video::SColor;

static irr::core::matrix4 toIrrMatrix(const aiMatrix4x4& assimpMatrix) {
    irr::core::matrix4 irrMatrix;

    irrMatrix[0] = assimpMatrix.a1;
    irrMatrix[1] = assimpMatrix.b1;
    irrMatrix[2] = assimpMatrix.c1;
    irrMatrix[3] = assimpMatrix.d1;

    irrMatrix[4] = assimpMatrix.a2;
    irrMatrix[5] = assimpMatrix.b2;
    irrMatrix[6] = assimpMatrix.c2;
    irrMatrix[7] = assimpMatrix.d2;

    irrMatrix[8] = assimpMatrix.a3;
    irrMatrix[9] = assimpMatrix.b3;
    irrMatrix[10] = assimpMatrix.c3;
    irrMatrix[11] = assimpMatrix.d3;

    irrMatrix[12] = assimpMatrix.a4;
    irrMatrix[13] = assimpMatrix.b4;
    irrMatrix[14] = assimpMatrix.c4;
    irrMatrix[15] = assimpMatrix.d4;

    return irrMatrix;
}

static SColor toIrrColor(const aiColor4D& color)
{
    return SColor(unsigned(color.a * 255), unsigned(color.r * 255),
                              unsigned(color.g * 255), unsigned(color.b * 255));
}

static irr::f32 getIrrFrame(double mTime, int frameOffset, double tps) {
    irr::f32 frame = float(mTime) + frameOffset;
    if (tps == 1)
        frame *= 60.f;
    return frame;
}

static const char* getBaseFileName(const char* sz) {
    auto p = strrchr(sz, '\\');
    return p ? p + 1 : sz;
}

static void createNode(SIrrMeshBuild* pid, const aiNode* node) {
    SIrrMeshBuild::Node sin;

    sin.name.assign(node->mName.C_Str());
    sin.nameParent.assign(node->mParent ? node->mParent->mName.C_Str() : "");
//    printf("CreateNode %s\n", node->mName.C_Str());

    sin.mtxLocal = toIrrMatrix(node->mTransformation);

    for (unsigned int i = 0; i < node->mNumMeshes; ++i) {
        sin.meshes.push_back(node->mMeshes[i]);
//        printf("added meshbuf[%d]\n", node->mMeshes[i]);
    }

    pid->nodes.push_back(sin);

    for (unsigned int i = 0; i < node->mNumChildren; ++i)
        createNode(pid, node->mChildren[i]);
}

bool readMeshFile(SIrrMeshBuild* pData, const char* szFileName, std::set<std::string>* txnames) {
    Assimp::Importer Importer;
    auto AssimpScene = Importer.ReadFile(szFileName, aiProcess_Triangulate | aiProcess_GenSmoothNormals | aiProcess_FlipUVs | aiProcess_MakeLeftHanded);
    if (!AssimpScene) {
        puts(Importer.GetErrorString());
        return NULL;
    }

    // Load nodes
    createNode(pData, AssimpScene->mRootNode);

    // Load meshes
    for (unsigned int i = 0; i < AssimpScene->mNumMeshes; ++i)
    {
        SIrrMeshBuild::MeshBuffer mb;
        auto paiMesh = AssimpScene->mMeshes[i];

        mb.verts.resize(paiMesh->mNumVertices);
        mb.bHasNormal = paiMesh->HasNormals();

        {
            mb.matType = irr::video::EMT_SOLID;

            auto mat = AssimpScene->mMaterials[paiMesh->mMaterialIndex];

            aiColor4D color;
            if(AI_SUCCESS == aiGetMaterialColor(mat, AI_MATKEY_COLOR_DIFFUSE, &color))
                mb.scDiffuse = toIrrColor(color);

            if(AI_SUCCESS == aiGetMaterialColor(mat, AI_MATKEY_COLOR_AMBIENT, &color))
                mb.scAmbient = toIrrColor(color);

            if(AI_SUCCESS == aiGetMaterialColor(mat, AI_MATKEY_COLOR_EMISSIVE, &color))
                mb.scEmissive = toIrrColor(color);

            if(AI_SUCCESS == aiGetMaterialColor(mat, AI_MATKEY_COLOR_SPECULAR, &color))
                mb.scSpecular = toIrrColor(color);

            if(AI_SUCCESS == aiGetMaterialColor(mat, AI_MATKEY_COLOR_TRANSPARENT, &color))
                mb.scTransparent = toIrrColor(color);
            else
                mb.scTransparent.set(0);

            float shininess = 0;
            if(AI_SUCCESS == aiGetMaterialFloat(mat, AI_MATKEY_SHININESS, &shininess))
                mb.fShininess = shininess;

            if (mat->GetTextureCount(aiTextureType_DIFFUSE) > 0) {
                aiString pathDiffuse;
                mat->GetTexture(aiTextureType_DIFFUSE, 0, &pathDiffuse);
                auto sz = getBaseFileName(pathDiffuse.C_Str());
                mb.txDiffuse.assign(sz);
                if(txnames) txnames->insert(sz);
            }
            if (mat->GetTextureCount(aiTextureType_NORMALS) > 0) {
                aiString pathNormal;
                mat->GetTexture(aiTextureType_NORMALS, 0, &pathNormal);
                auto sz = getBaseFileName(pathNormal.C_Str());
                mb.txNormal.assign(sz);
                mb.matType = irr::video::EMT_PARALLAX_MAP_SOLID;
                if(txnames) txnames->insert(sz);
            }
        }

        auto defaultDiffuse = mb.txDiffuse.length() > 0 ? SColor(255, 255, 255, 255) : mb.scDiffuse;

        for (unsigned int j = 0; j < paiMesh->mNumVertices; ++j)
        {
            const aiVector3D vertex = paiMesh->mVertices[j];
            mb.verts[j].Pos = vector3df(vertex.x, vertex.y, vertex.z);

            if (paiMesh->HasNormals()) {
                const aiVector3D normal = paiMesh->mNormals[j];
                mb.verts[j].Normal = vector3df(normal.x, normal.y, normal.z);
            }

            if (paiMesh->HasVertexColors(0)) {
                const aiColor4D color = paiMesh->mColors[0][j];
                mb.verts[j].Color = toIrrColor(color);
            }
            else {
                mb.verts[j].Color = defaultDiffuse;
            }

            if (paiMesh->GetNumUVChannels() > 0) {
                const aiVector3D uv = paiMesh->mTextureCoords[0][j];
                mb.verts[j].TCoords = vector2df(uv.x, uv.y);
            }
        }
        if (paiMesh->HasTangentsAndBitangents()) {
            mb.Tangents.resize(paiMesh->mNumVertices);
            mb.Binormals.resize(paiMesh->mNumVertices);
            for (unsigned int j = 0; j < paiMesh->mNumVertices; ++j) {
                aiVector3D tangent = paiMesh->mTangents[j];
                aiVector3D bitangent = paiMesh->mBitangents[j];
                mb.Tangents[i]  = vector3df(tangent.x, tangent.y, tangent.z);
                mb.Binormals[i] = vector3df(bitangent.x, bitangent.y, bitangent.z);
            }
        }

        if (paiMesh->GetNumUVChannels() > 1) { // It has 2nd UV channel.
            mb.TCoords2.resize(paiMesh->mNumVertices);
            for (unsigned int j = 0; j < paiMesh->mNumVertices; ++j) {
                const aiVector3D uv = paiMesh->mTextureCoords[0][j];
                mb.TCoords2[j] = vector2df(uv.x, uv.y);
            }
        }

        mb.indices.resize(paiMesh->mNumFaces * 3);

        for (unsigned int j = 0; j < paiMesh->mNumFaces; ++j) {
            const aiFace face = paiMesh->mFaces[j];
#if 0
            mb.indices[3 * j    ] = face.mIndices[0];
            mb.indices[3 * j + 1] = face.mIndices[1];
            mb.indices[3 * j + 2] = face.mIndices[2];
#else
            mb.indices[3 * j    ] = face.mIndices[2];
            mb.indices[3 * j + 1] = face.mIndices[1];
            mb.indices[3 * j + 2] = face.mIndices[0];
#endif
        }

        // Skinning
        mb.bones.resize(paiMesh->mNumBones);

        for (unsigned int j = 0; j < paiMesh->mNumBones; ++j) {
            auto bone = paiMesh->mBones[j];
            mb.bones[j].name.assign(bone->mName.C_Str());

            mb.bones[j].weights.resize(bone->mNumWeights);
            for (unsigned int h = 0; h < bone->mNumWeights; ++h)
            {
                aiVertexWeight weight = bone->mWeights[h];
                mb.bones[j].weights[h].weight   = weight.mWeight;
                mb.bones[j].weights[h].idVertex = weight.mVertexId;
            }
            mb.bones[j].mtxOffset = toIrrMatrix(bone->mOffsetMatrix);
        }
        pData->meshbufs.push_back(mb);
    }

    // Load animations
    int frameOffset = 0;
    pData->animations.resize(AssimpScene->mNumAnimations);

    for (unsigned int i = 0; i < AssimpScene->mNumAnimations; ++i) {
        auto anim = AssimpScene->mAnimations[i];
        auto& ida = pData->animations[i];

        if (anim->mTicksPerSecond != 0.f)
            ida.speed = (irr::f32)anim->mTicksPerSecond;
        if (anim->mTicksPerSecond == 1)
            ida.speed = 60.f;

        ida.channels.resize(anim->mNumChannels);

        for (unsigned int j = 0; j < anim->mNumChannels; ++j) {
            aiNodeAnim* nodeAnim = anim->mChannels[j];
            auto& idac = ida.channels[j];
            idac.nodeName.assign(nodeAnim->mNodeName.C_Str());

            idac.posKeys.resize(nodeAnim->mNumPositionKeys);
            for (unsigned int k = 0; k < nodeAnim->mNumPositionKeys; ++k) {
                auto& idk = idac.posKeys[k];
                aiVectorKey key = nodeAnim->mPositionKeys[k];

                idk.frame = getIrrFrame(key.mTime, frameOffset, anim->mTicksPerSecond);
                idk.val = vector3df(key.mValue.x, key.mValue.y, key.mValue.z);
            }

            idac.rotationKeys.resize(nodeAnim->mNumRotationKeys);
            for (unsigned int k = 0; k < nodeAnim->mNumRotationKeys; ++k) {
                auto& idk = idac.rotationKeys[k];
                aiQuatKey key = nodeAnim->mRotationKeys[k];
                aiQuaternion assimpQuat = key.mValue;

                irr::core::quaternion quat(-assimpQuat.x, -assimpQuat.y, -assimpQuat.z, assimpQuat.w);
                quat.normalize();

                idk.frame = getIrrFrame(key.mTime, frameOffset, anim->mTicksPerSecond);
                idk.val = quat;
            }
            
            idac.scaleKeys.resize(nodeAnim->mNumScalingKeys);
            for (unsigned int k = 0; k < nodeAnim->mNumScalingKeys; ++k) {
                auto& idk = idac.scaleKeys[k];
                aiVectorKey key = nodeAnim->mScalingKeys[k];

                idk.frame = getIrrFrame(key.mTime, frameOffset, anim->mTicksPerSecond);
                idk.val = vector3df(key.mValue.x, key.mValue.y, key.mValue.z);
            }
        }
        frameOffset += int(anim->mDuration);
    }
    return true;
}
